/*
 * log_dump: dump current system states
 *
 * Copyright (c) 2016, 2018 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include <assert.h>
#include <dirent.h>
#include <fcntl.h>
#include <getopt.h>
#include <libgen.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/stat.h>
#include <time.h>
#include <unistd.h>

#include <system_info.h>
#include <tzplatform_config.h>

#include "defs.h"
#include "dbus-handler.h"
#include "log_dump.h"
#include "shared/config.h"
#include "shared/spawn.h"
#include "shared/util.h"

#define SYSTEM_INFO_KEY_BUILD_STRING "http://tizen.org/system/build.string"
#define DIR_UMASK 0022

static const struct option opts[] = {
	{ "normal", no_argument, 0, OPT_NORMAL },
	{ "short",  no_argument, 0, OPT_SHORT  },
	{ "dbus",   no_argument, 0, OPT_DBUS   },
	{ 0, 0, 0, 0 }
};

/* crash worker configuration - for admin-selectable CrashRootPath */
config_t config;
/* tzplaform vars */
char *dir_scripts;
/* dynamic vars */
char *dir_dump;
char *dir_log;
char *dir_debug;
char *dir_temp;  // temp rootdir
char *dir_temp_logdump; // rootdir for this log_dump invocation
char *version_string;
/* timestamp */
const char timestamp_format[] = "%Y%m%d%H%M%S";
char timestamp_string[20]; /* as per format above */

static bool init_temp_dir(char *const temp_root, char **temp_dir)
{
	assert(temp_root);
	assert(temp_dir);

	char *template = NULL, *path = NULL;
	if (asprintf(&template, "%s/log.XXXXXX", temp_root) > 0)
		path = mkdtemp(template);

	if (!path) {
		_E("Unable to create temporary directory at mkdtemp(%s): %m", template);
		free(template);
		return false;
	}

	*temp_dir = path;
	return true;
}

static char *crash_root_get(void)
{
	return config.crash_root_path;
}

static bool init_vars(const char *crash_root)
{
	if (!crash_root)
		return false;

	if (asprintf(&dir_log, "%s/log", crash_root) <= 0
	    || asprintf(&dir_debug, "%s/debug", crash_root) <= 0
	    || asprintf(&dir_dump, "%s/dump", crash_root) <= 0
	    || asprintf(&dir_temp, "%s/temp", crash_root) <= 0)
		goto fail;

	make_dir(crash_root, "temp", DIR_UMASK);
	if (!init_temp_dir(dir_temp, &dir_temp_logdump))
		goto fail;

	make_dir(dir_temp_logdump, "log", DIR_UMASK);
	make_dir(crash_root, "debug", DIR_UMASK);

	_D("config: dir_log is %s", dir_log);
	_D("config: dir_dump is %s", dir_dump);
	_D("config: dir_debug is %s", dir_debug);
	_D("config: dir_temp is %s", dir_temp);
	_D("config: dir_temp_logdump is %s", dir_temp_logdump);

	dir_scripts = strdup(tzplatform_getenv(TZ_SYS_DUMPGEN));
	_D("config: dir_scripts is %s", dir_scripts);

	if (system_info_get_platform_string(SYSTEM_INFO_KEY_BUILD_STRING, &version_string) != SYSTEM_INFO_ERROR_NONE) {
		_W("Failed to system_info_get_platform_string for " SYSTEM_INFO_KEY_BUILD_STRING);
		version_string = NULL;
	}
	_D("version_string is %s", version_string);

	time_t cur_time;
	struct tm loc_tm;
	cur_time = time(NULL);
	localtime_r(&cur_time, &loc_tm);
	strftime(timestamp_string, sizeof(timestamp_string), timestamp_format, &loc_tm);
	_D("timestamp_string is %s", timestamp_string);

	assert(dir_log);
	assert(dir_dump);
	assert(dir_debug);
	assert(dir_temp);
	assert(dir_temp_logdump);
	return true;

fail:
	free(dir_log);
	free(dir_dump);
	free(dir_debug);
	free(dir_temp);
	free(dir_temp_logdump);
	return false;
}

static void usage(void)
{
	printf("Usage: log_dump {OPTION}\n"
	       "Options:\n"
	       "  --normal    dump all logs (uses scripts from %s)\n"
	       "  --short     dump_systemstate logs only\n"
	       "  --dbus      become dbus service\n",
	       dir_scripts);
}

static bool dump_scripts(char *const workdir, char *const scriptsdir)
{
	struct dirent **dir_list = NULL;
	char command[PATH_MAX];
	int script_num, i;

	script_num = scandir(scriptsdir, &dir_list, NULL, NULL);
	if (script_num < 0) {
		_E("scandir %s: %m", scriptsdir);
		return false;
	}

	for (i = 0; i < script_num; i++) {
		const char *const name = dir_list[i]->d_name;
		int type = dir_list[i]->d_type;

		if (strcmp(name, ".") == 0 || strcmp(name, "..") == 0)
			continue;

		if (type != DT_REG) {
			_D("Ignoring: not a regular file: %s", name);
			continue;
		}
		snprintf(command, sizeof(command), "%s/%s", scriptsdir, name);
		if (access(command, X_OK) != 0) {
			_W("Ignoring: file not executable: %s", command);
			continue;
		}

		_D("Calling scriptlet: %s", command);

		char *const av[] = {command, workdir, NULL};
		(void)spawn_wait(av, NULL, NULL, 0, NULL);
	}

	for (i = 0; i < script_num; i++)
		free(dir_list[i]);
	free(dir_list);

	return true;
}

static bool dump_systemstate(const char *const destdir, const char *const timestr, int *exit_code)
{
	char *dump_path = NULL;

	if (asprintf(&dump_path, "%s/log/dump_systemstate_%s.log", destdir, timestr) < 0) {
		_E("asprintf: %m");
		return false;
	}

	char *av[] = {DUMP_SYSTEMSTATE_BIN_PATH, "-k", "-d", "-j", "-e", "-f", dump_path, NULL};
	spawn_param_s param = { .fn = spawn_setstdout, .u.int_val = STDERR_FILENO };
	bool is_ok = spawn_wait(av, NULL, &param, 0, exit_code);

	free(dump_path);

	return is_ok;
}

static bool compress(char *const destdir, char *const tempdir, char *const versionstr, char *const timestr, int *exit_code)
{
	char *archive_path = NULL;

	if (asprintf(&archive_path, "%s/log_dump_%s%s.zip", destdir, versionstr ?: "", timestr) < 0) {
		_E("asprintf: %m");
		return false;
	}

	_D("compress tempdir is %s", tempdir);
	char *av[] = {"/bin/zip", "-qyr", archive_path, ".", NULL};
	spawn_param_s param1 = { .fn = spawn_nullstdfds };
	spawn_param_s param0 = { .fn = spawn_chdir, .u.char_ptr = tempdir, .next = &param1 };
	bool is_ok = spawn_wait(av, NULL, &param0, 0, exit_code);

	_I("Storing report at %s", archive_path);

	fsync_path(archive_path);
	free(archive_path);

	return is_ok;
}

int log_dump(int option)
{
	broadcast_logdump_start();

	int ret = -1;

	if (!dump_systemstate(dir_temp_logdump, timestamp_string, NULL))
		goto out;

	if (option == OPT_NORMAL)
		(void)dump_scripts(dir_temp_logdump, dir_scripts);

	compress(dir_debug, dir_temp_logdump, version_string, timestamp_string, NULL);

	/* cleanup */
	ret = remove_dir(dir_temp_logdump, 1);
	if (ret < 0)
		_W("Failed to delete dump directory at %s", dir_temp_logdump);

	ret = 0;
out:
	broadcast_logdump_finish();

	return ret;
}

int delete_dump(void)
{
	_D("delete_dump called");

	remove_dir(dir_log, 0);
	remove_dir(dir_debug, 1);
	remove_dir(dir_dump, 0);
	remove_dir(dir_temp, 0);

	return 0;
}

static void die(void)
{
	usage();
	exit(EXIT_FAILURE);
}

int main(int argc, char *argv[])
{
	int c, ret;
	int option;

	if (!config_init(&config, CRASH_MANAGER_CONFIG_PATH))
		return false;

	/* need to do this first, because even usage() uses the vars */
	if (!init_vars(crash_root_get())) {
		printf("Unable to initialize - please check program logs");
		exit(EXIT_FAILURE);
	}

	if (argc < 2)
		die();

	option = -1;
	while ((c = getopt_long_only(argc, argv, "", opts, NULL)) != -1) {
		if (option >= 0 || c < _OPT_MIN || c > _OPT_MAX)
			die();
		option = c;
	}

	if (option == OPT_DBUS)
		ret = log_dump_dbus();
	else
		ret = log_dump(option);

	config_free(&config);

	return ret < 0 ? EXIT_FAILURE : EXIT_SUCCESS;
}
